import.grid<-function(gridfile = "c:/research/papers/extremewinds2/gridboxes.txt",prefix="Region")
  {
    grid<-read.table(header=F,gridfile)
    colnames(grid)<-c("lon","lat")
    na<-c(0,which(is.na(grid$lon)),nrow(grid)+1)
    lna<-length(na)-1
    out<-vector(mode="list",lna)
    for(i in 1:lna)
      {
        ids<-c(na[i]+1,na[i+1]-1)
        if(ids[2]<ids[1]) outi<-NULL
        else
        outi<-grid[ids[1]:ids[2],]
        out[[i]]<-outi
      }
    names(out)<-paste(prefix,1:lna,sep=".")
    out
  }

add.grid<-function(data=best.use,grid=grid)
  {
  ###modified to remove Sid information since we are no longer using by storm id, it is easy to recreate.
    x<-data[,c("lon","lat")]
    lgrid<-length(grid)
    grid.ind<-vector(mode="list",lgrid)
    names(grid.ind)<-names(grid)
    for(i in 1:lgrid)
      {
        grid.ind[[i]]<-is.land(grid=grid[[i]],points=x,center=NULL, method = "l") 
      }
    grid.df<-do.call("data.frame",grid.ind)
    data<-cbind(data,grid.df)
    data
  }





"is.land"<-function(grid=landGrid,points,center=NULL,method="l")
{
###This function determines whether a vector of points is inside hull
###by considering the number of rotations a point on the curve makes about a single point.
###The result counts is this number it should be the number of counter clockwise rotations.
###The sign denotes the curve direction (+ couterclockwise) about an interior point.
###This function was converted from a C program
###The center is chosen so that the grid comprises a simple closed curve
###The result is a vector with the number of track crossings
###This works for all closed curves
###Method=1 use lapply, slower but works on large data sets
###Method=2 use outer3, faster, but works only on single tracks.
  if(!is.null(center))
    grid<-rbind(center,grid)
  xp<-grid[,1]
  yp<-grid[,2]
  x<-points[,1]
  y<-points[,2]
  if(any(colnames(points)[1:2] != colnames(grid)[1:2]))
    warning(paste("points column names:",colnames(points)[1:2],"are not equal to","grid column names:",colnames(grid)[1:2]))
  lx<-length(xp)
  ##Do not add last point, remove first point instead assume that each loop is closed by having the first and last points
  ##duplicated. (use close.curves)
  xpi<-xp[-1]
  xpj<-xp[-lx]
  ypi<-yp[-1]
  ypj<-yp[-lx]
  gridcross<-cbind(xpi,xpj,ypi,ypj)
  xy<-cbind(x,y)
  ##get x crossings
  if(method=="l")
    {
  crossings<-unlist(lapply(1:nrow(xy),
                           function(i,xy,grid)
                           {
                             x<-xy[i,1]
                             y<-xy[i,2]
                             xpi<-grid[,1]
                             xpj<-grid[,2]
                             xi<-x < xpi
                             xj<-x < xpj
### xcross=(xj&(!xi)) | (xi&(!xj) )  ==
                             vi<-(xj-xi)
                             vi[is.na(vi)]<-0 ###Closed Curves separated by NAs
                             vil<-as.logical(vi)
                             if(!any(vil)) return(0)
                             vi<-vi[vil]
                             grid<-grid[vil,]
                             xpi<-grid[,1]     
                             xpj<-grid[,2]
                             ypi<-grid[,3]
                             ypj<-grid[,4]
                             xdelta<-xpi-xpj
                             ydelta<-ypi-ypj
                             sum(vi*(y<(ypi+(x-xpi)*ydelta/xdelta)))
                           },xy=xy,grid=gridcross))
  is.inside<-as.logical(crossings)
}
else
  {##Keep for later, may be faster
  ydelta<-ypj-ypi
  xdelta<-xpj-xpi
   gridcross<-cbind(xdelta,ydelta,xpi,ypi)
   xj<-outer(x,xpj,"<")
   xi<-outer(x,xpi,"<")
  ## xcross=(xj&(!xi)) | (xi&(!xj) )  ==
  xcross<-(xj-xi)
  xcross[is.na(xcross)]<-0 ###Closed Curves separated by NAs
   crossings<-outer3(x=xy,y=gridcross,fun=function(xy,grid,v)
                    {
                      v<-as.vector(v)
                      x<-xy[,1]
                      y<-xy[,2]
                      xdelta<-grid[,1]
                      ydelta<-grid[,2]
                      xp<-grid[,3]
                      yp<-grid[,4]
                      ifelse(v,v*(y<(yp+(x-xp)*ydelta/xdelta)),0)
                    },v=xcross
                    )
  is.inside<-as.logical(rowSums(crossings))
}
#  print(all(is.insideo==is.inside))
  return(is.inside)
}
##C code:
##  int pnpoly(int npol, float *xp, float *yp, float x, float y)
##    {
##      int i, j, c = 0;
##for (i = 0, j = npol-1; i < npol; j = i++) {
##  if ((((xp[i]<=x) && (x<xp[j])) ||
##       ((xp[j]<=x) && (x<xp[i]))) &&
#      
###/*where is the y crossing value?*/
##                   
##      (y < (yp[j] - yp[i]) * (x - xp[i]) / (xp[j] - xp[i]) + yp[i]))
##          c = !c;
##      }
##      return c;
##    }
