##These functions augment the ismev routines.

###pp.rate calculates the rate given gev parameters mu=x[1],sigma=x[2],xi=x[3]
###pp.rate for y=threshold, is the threshold crossing rate
"pp.rate" <-function(x,y)
{
denom<-1+ x[3]*(y-x[1])/x[2]
ifelse(denom<=0,0,denom^(-1/x[3]))
}

"pp.rl"<-function(x,y)  
{
    lambda<-1/y
    if (x[3] != 0) 
        x[1] + x[2]/x[3]* (lambda^(-x[3]) - 1)
    else x[1]-x[2]*log(lambda)
}

###pp.sigma, calculates the GPD scale parameter for a vector of thresholds y
###As y increases 
"pp.sigma"<-function(x,y){x[2]+x[3]*(y-x[1])}

###This is for completeness
"pp.xi"<-function(x,y){ rep(x[3],length(y))} 

"derivparam" <-
function(x,f,delta=10e-6,y=1){
lx<-length(x)
xp<-xn<-x0<-array(rep(x,each=lx),c(lx,lx))
xd<-row(x0)==col(x0)
xp[xd]<-x0[xd]+delta/2
xn[xd]<-x0[xd]-delta/2
yp<-apply(xp,1,f,y)
yn<-apply(xn,1,f,y)
(yp-yn)/delta
}

get.conf<-function(y,model,p=c(.05,.10,.5,.90,.95),f=c("pp.rate","pp.rl"),digits=0,units="kt")
{
###y is the return level of interest
###returns confidence interval on return period for storms of a given strength or stronger.
f<-match.arg(f)
ff<-get(f)
p0<-ff(x=model$coef,y=y)
sigma<-get.sigma(mle=model$coef,cov=model$cov,y=y,f=ff)
nr<-length(p0)
nc<-length(p)
if(f=="pp.rate")
{
q<-round(1/qnorm(rep(1-p,each=nr),mean=rep(p0,nc),sd=rep(sigma,nc)),digits=digits)  ###use 1/return rate, don't use delta on return period
q<-ifelse(q<0,Inf,q)
}
else
q<-round(qnorm(rep(p,each=nr),mean=rep(p0,nc),sd=rep(sigma,nc)),digits=digits)
dim(q)<-c(nr,nc)
colnames(q)<-paste(round(p,3)*100,"%",sep="")
rownames(q)<-paste(format(y),units,sep=" ")
return(q)
}


get.sigma<-function(mle,cov,y,f){dd<-derivparam(x=mle,f=f,y=y); return(sqrt(apply(dd,1,q.form,m=cov)))}  


"np.fit"<-function(x, rate)
{
    return(1 - exp( - rate * (1 - ((rank(x) - 0.5)/length(x)))))
}

"pp.pois.rl" <- function(a, mat = NULL, dat, rate = 1, minq = 0, addtext = "for 29N 89W", plot.lower = NULL, plot.upper = NULL, rp = NULL, ...)
{
    ###
    ### function called by exceed.pp.plot, called in turn by exceed.model.plot, which is called by exceed.model.climates,
        ### used in exceed.ssc. See exceed.pp.plot for details.
    ### produces return level curve and 95 % confidence intervals, modified to handle return periods 1,2,...9,10,20,...90,100,200,..,1000
        ### 
        ### on usual scale
    ### rps, return periods of interest 
    eps <- 9.9999999999999995e-007
    a1 <- a
    a2 <- a
    a3 <- a
    a1[1] <- a[1] + eps
    a2[2] <- a[2] + eps
    a3[3] <- a[3] + eps
    if(is.null(rp))
        rp <- c(seq(1, 9, by = 1), seq(10, 90, by = 10), seq(100, 1000, by = 100))
    f <- exp(-1/rp)   ###return period=1/return rate. The return period as the number of years between two events is 1/(1-f) and its about .5 years larger than rp. This matches normal definitions.
    #   f <- c(seq(0.01, 0.09, by = 0.01), 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99, 0.995, 0.999)
    q <- gevq(a, 1 - f)
    keep.q <- q >= minq
    f <- f[keep.q]
    q <- q[keep.q]
    d1 <- (gevq(a1, 1 - f) - q)/eps
    d2 <- (gevq(a2, 1 - f) - q)/eps
    d3 <- (gevq(a3, 1 - f) - q)/eps
    d <- cbind(d1, d2, d3)
    if(is.null(mat))
        v <- 0
    else v <- apply(d, 1, q.form, m = mat)
    upper <- q + 1.96 * sqrt(v)
    lower <- 2 * q - upper        
    rp <- -1/log(f)
    if(!is.null(plot.lower))
        ymin <- plot.lower
    else ymin <- min(dat, lower)
    if(!is.null(plot.upper))
        ymax <- plot.upper
    else ymax <- max(dat, upper)
    plot(rp, q, log = "x", type = "n", xlim = c(1, 1000), ylim = c(ymin, ymax), xlab = "Return Period", ylab = 
        "Return Level")
    title(paste("Return Level Plot", addtext, sep = "\n"))
    lines(rp, q, ...)
    lines(rp, upper, col = 4, ...)
    lines(rp, lower, col = 4, ...)
    rp.emp <- -1/log(1 - np.fit(dat, rate = rate))
    rl <- dat
    points(rp.emp, dat,err=-1,pch=".",cex=3,col="red")
    return(list(return.level.curves = data.frame(rp = rp, lower = lower, q = q, upper = upper), empirical.plot = data.frame(
        rp = rp.emp, rl = rl), model = list(coef = a, cov = mat)))
}

exceed.pp.plot<-function(wind,npy=1,threshold=34,addtext="Return Level Plot",xunit="Years",yunit="kt",...)
   ##called by exceed.model.plot, documentation provided as needed.
{
  model.fit<-try(pp.fit(wind,npy=npy,threshold=threshold,cov=T,show=F))
  if(class(model.fit)=="Error")
    model.fit<-try(pp.fit(wind,npy=npy,threshold=threshold,cov=F,show=F))
  if(class(model.fit)=="Error")
    {
      cat(paste("No model fit for", addtext,".\n"))
      return(NULL)
    }
  wind.data<-pp.pois.rl(model.fit$mle,model.fit$cov,dat=wind,rate=npy,addtext=addtext,...)
  mtext(side=1,line=2,xunit)
  mtext(side=2,line=2,yunit)
  abline(h=threshold,col=6)
  return(wind.data)
}
###The functions gpd.to.gev,  grad, gpd.to.gev.cov, gpd.fit.gev, were created to convert stable fits with
###using gpd.fit to a GEV points over threshold fit, under the assumption that the threshold is constant.
###These fucntions work only when sigma, xi, rate, are not dependent on covariates.
###Provided "as is" without any documentation. 
"gpd.to.gev" <- function(parameters = NULL, sigma, xi, rate, thresh)
{
  if(!is.null(parameters))
    {
      rate <- parameters[1]
      sigma <- parameters[2]
      xi <- parameters[3]
    }
  if(xi == 0)
    mu = sigma * log(rate) + thresh
  else mu <- thresh + sigma/xi * (rate^( xi)- 1)
  sigma <- sigma + xi * (mu - thresh)
  return(c(mu = mu, sigma = sigma, xi = xi))
}

"grad" <- function(f, x, ep = 10^-5, ...)
{
###This function performs a gradient derivative, where f(x) returns a vector, then grad(f(x)) is a matrix.
###Note that one column per elemnt of x and one row per element of f(x) so that delta(f(x)) = grad(f,x) %*% deltax
###where delta x is a column 
## if y=f(x)  then grad(f)[i,j =  delta(y[i])/partial(x[j])
  eps <- ep * x
  n <- length(x)
  x0 <- f(x, ...)
  m <- matrix(0, length(x0), n)
  colnames(m) <- names(x)
  rownames(m) <- names(x0)
  for(i in 1:n)
    {
      x1 <- x
      x1[i] <- x[i] + eps[i]
      x2 <- x
      x2[i] <- x[i] - eps[i]
      m[, i] <- (f(x1, ...) - f(x2, ...))/(2 * eps[i])
    }
  m
}

"gpd.to.gev.cov" <- function(parameters, gpd.cov, thresh, time)
{
  grad.gpd.gev <- grad(gpd.to.gev, parameters, thresh = thresh)
  sigma <- matrix(0, 3, 3)
  sigma[2:3, 2:3] <- gpd.cov
  sigma[1, 1] <- parameters[1]/time
  return(grad.gpd.gev %*% sigma %*% t(grad.gpd.gev))
}
gpd.fit.gev<-function(x,thresh,time)
  {
    
    ##This assumes that the count per time  is poission with lambda = rate)
    xt<-x[x>thresh]
    rate<-length(xt)/time 
    var.rate<-rate/time   #assume time known xt poisson, then var(xt/time)= var(xt)/time^2 approx n/time^2 = rate/time 
    gpd.dat<-gpd.fit(x,ufun=thresh)
    gpd.mle<-c(rate,gpd.dat$mle)
    names(gpd.mle)<-c("rate","sigma","xi")
    gpd.cov<-gpd.dat$cov
    gpd.cov.out<-matrix(0,3,3)
    gpd.cov.out[2:3,2:3]<-gpd.cov
    gpd.cov.out[1,1]<-var.rate
    dimnames(gpd.cov.out)<-rep(list(c("rate","sigma","xi")),2)
    gpd.mlev<-as.vector(gpd.mle)
    pp.mle<-gpd.to.gev(parameters=gpd.mlev,thresh=thresh)
    pp.cov<-gpd.to.gev.cov(parameters=gpd.mlev,gpd.cov=gpd.cov,thresh=thresh,time=time)
    return(list(gpd.mle=gpd.mle,gpd.cov=gpd.cov.out,pp.mle=pp.mle,pp.cov=pp.cov))
  }
