####Utility Routines for HurRiskMain.R
###get.max.flag could be incorporated into get.max as well...
###We use the x11() device, you can set this up on th Mac by installing the application from the installation disk for Max OS X Version 10.4.11 (Have not tested Leopard yet)
get.max.flags<-function(x=best.use,fields=c("WmaxS","DWmaxDt"),flag="Region.4",se=c(1851,2006),field.suffix=c("",".d"),base.fields=c("SYear","Sid","name"),keep.fields=c("Yr","Mo","Da","hr","lon","lat","Wmax","WmaxS","DWmaxDt","Type","Shour","maguv","diruv"))
{
xx<-rep(list(NULL),length(fields))
lf<-length(fields)
if(length(field.suffix) != lf) field.suffix=c(letters[1:lf])
if(!is.null(flag) && flag!="") base.fields<-c(base.fields,flag)  ###flag is base field, should all be true in result
x<-x[,unique(c(base.fields,fields,keep.fields))]  ##keep all fields order fields
bm<-match(colnames(x),base.fields,nomatch=0)      ##which fields are base fields, that we only need once
for(i in 1:lf)
{
  tx<-get.max.flag(x=x,field=fields[i],flag=flag,se=se)
  if(i>1) tx<-tx[-bm] ###remove base fields
  colnames(tx)<-paste(colnames(tx),field.suffix[i],sep="")
  xx[[i]]<-tx
}
out<-do.call("cbind",xx)
if(!is.null(flag) && flag!="") out[flag]<-NULL
out
}

###select Maximum Winds, and Wind Speed at maximum intensification.
get.max.landfall<-function(x=best.use,field="Wmax",flag="Region.4",se=c(1851,2006))
{
###This function derives a new data set consisting of
#1. Wmax in Basin, 
pos<-split(1:nrow(x),x$Sid)   
xmax<-sapply(pos,function(i,xx,ff)
        {
            xi<-xx[i,ff] 
            xf<-xx[i,flag]
            xM<-xx[i,M]
            i[which(xi==max(xi))[1]]
        },
        xx=x,ff=field)         
    out<-x[xmax,] 
    return(out)     
}

get.max.flag<-function(x=best.use,field="Wmax",flag="Region.4",se=c(1851,2006))
{
###data has field flag that is used to subselect data for maximization.
###data is not converted per search radius
if(is.null(flag) || flag=="") flag<-"T"
x<-subset(x,SYear>=se[1] & SYear<=se[2] & eval(parse(text=flag)))
if(nrow(x)>0)
{
    pos<-split(1:nrow(x),x$Sid)   
    xmax<-sapply(pos,function(i,xx,ff)
        {
            xi<-xx[i,ff] 
            i[which(xi==max(xi))[1]]
        },
        xx=x,ff=field)         
    out<-x[xmax,] 
    return(out)     
}
else
return(x)
}

get.max<-function(lon,lat,radius,distunits=c("nm","miles","km"),velunits=c("kt","mph","k/h","m/s"),x=best.use,field="WmaxS",umin=35,se=se)
{
###Would like to add a weight function, currently we just get the 
###lat, lon and radius are point values that we use to collect information needed to generate the extreme value plots.
###radius is in nautical miles
###results in kts.
###for simplicity umin is given in kt, can change later
###We can change as needed.
###Convert radius to units

    convs<-c(1,0.868976242,0.539956803, 1.943844492)   ###convert nm to nm, miles to nm, km to nm, m/s to kt
    convdistin<-match.unit(distunits,conv=convs[1:3]) 
    convdistout<-1/convdistin
    radius<-radius*convdistin ###from here on out units in nm...
    convelout<-1/match.unit(velunits,convs)  ###convert kt to kt, mph, k/h, m/s

    xin<-inside.lonlat(x=x,lon=lon,lat=lat,yrlim=se,r=radius) & x[,field] > umin
    if(sum(xin)==0) stop("Please increase your search radius, no storms selected")
    x<-x[xin,]
    pos<-lonlattoxyz(data=x)
    pos0<-lonlattoxyz(lon=lon,lat=lat)
    xdist<-distance.from.center(pos0=pos0,pos=pos)
    xin<- xdist <= radius
    xdist<-xdist[xin]
    if(sum(xin)==0) stop("Please increase your search radius, no storms selected")
    x<-x[xin,]
    pos<-split(1:nrow(x),x$Sid)   

    xmax<-sapply(pos,function(i,xx,ff)
        {
            xi<-xx[i,ff] 
            i[which(xi==max(xi))[1]]
        },
        xx=x,ff=field)

 
    out<-x[xmax,]  
    xdist<-xdist[xmax]
    return(cbind(out,y=out[,field]*convelout,dist=xdist*convdistout))
}

inside.box<-function(lon,lat,yr=NULL,lonlim,latlim,yrlim=NULL)
{
if(is.null(yr)){ yr=0 ;yrlim=c(0,0) }
###this function generates a box xlim, ylim
###this fucntion returns a logical vector for each position within the lon,lat, box. 
    return(lon>=lonlim[1] & lon<lonlim[2] & lat>=latlim[1] & lat<latlim[2] & yr >= yrlim[1] & yr <= yrlim[2] )
}

distance.from.center<-function(pos0,pos)
{
    CX<-pos0[1,1]
    CY<-pos0[1,2]
    CZ<-pos0[1,3]
    X<-pos[,1]-CX
    Y<-pos[,2]-CY
    Z<-pos[,3]-CZ
    return(sqrt(X^2+Y^2+Z^2))
}

inside.lonlat<-function(x,lat,lon,yrlim=NULL,r,r0=3440.07)
{
###x is a dataframe or matrix with columns "lon" and "lat",
###r is the radius, r0 is the radius of the earth
###returns a subset of xx 

###In the merdinal direction r=r0*diff(latlim)/2*pi/180 => 
    difflatlim<-c(-1,1)*r/r0*180/pi
    latlim<-difflatlim+lat
    if(is.null(yrlim)) 
        yr<-NULL
    else 
        yr=x[,"SYear"]
###In zonal direction r=r0*diff(lonlim)/2*pi*cos(lat)/180 
###but cos(lat) changes! since we want r smaller than actual r
###use min(cos(lat)) or min(cos(latlim)) in formula
    lonlim<-lon+c(-1,1)*r/r0 * 180/pi /min(cos(latlim/180*pi))
    xx<-inside.box(lon=x[,"lon"],lat=x[,"lat"],yr=yr,lonlim=lonlim,latlim=latlim,yrlim=yrlim)
    return(xx)
}

"lonlattoxyz" <- function(lon=NULL, lat=NULL, r = 3440.07,data=NULL ) #3440.07 knots
{
  #This function converts longitude and lattitude to x,y,z coordiantes. Height is zero.
  #Input: lon, lat, vectors or data, data frame with lon lat columns,
  #       r is the radius of earch as 3440.07 nautical miles,
  #returns matrix with columns x,y,z
  if(!is.null(data))
    {
      lon<-data[,"lon",drop=T]
      lat<-data[,"lat",drop=T]
    }
    rbase <- r * cos((lat * pi)/180)
    cbind(x = rbase * cos((lon * pi)/180), y = rbase * sin((lon *
        pi)/180), z = r * sin((lat * pi)/180))
}


"xyztolonlatheight"<-function(x=NULL,y=NULL,z=NULL,data=NULL,r0=3440.07) #3440.07 nautical miles, change for other units
  {
    #This function is used many places to convert xyz coordintes to lonlatheight,
    #inputs: x,y,z equal length vectors of x,y,z coordinates, if data argument is given will use data$x, data$y, data$z
    #r0 is the radius of the earth, here given as 3440.07 kt (mean).
    #returns matrix with lon, lat, and height columns
    if(!is.null(data))
      {
        x<-data[,"x",drop=T]
        y<-data[,"y",drop=T]
        z<-data[,"z",drop=T]
      }
    R<-sqrt(x^2+y^2+z^2)
    degtorad<-180/pi
    lat<-degtorad*asin(z/R)
    lon<-degtorad*atan2(y,x)
    height<-R-r0
    return(cbind(lon=lon,lat=lat,height=height))
  } 

cross<-function(x,y)
{
###standard cross product
out<-c(x[2]*y[3]-x[3]*y[2],-(x[1]*y[3]-y[1]*x[3]),x[1]*y[2]-y[1]*x[2])
out
}

match.unit<-function (arg, conv, choices ) 
{
    if (missing(choices)) 
    {
        formal.args <- formals(sys.function(sys.parent()))
        choices <- eval(formal.args[[deparse(substitute(arg))]])
    }
    if(is.null(arg))
        arg<-choices[1]
    arg<-tolower(arg)
    i <- pmatch(arg, choices)[1]
    if (any(is.na(i))) 
        stop("'arg' should be one of ", paste(choices, collapse = ", "))
    conv[i]
}

select.location<-function(txt,colnames=c("lon","lat"),color=color,digits=0)
{
title(txt,col.main=color)
loc<-locator(1)
if(!is.null(digits)) loc<-lapply(loc,round,digits=digits)
points(loc$x,loc$y,pch="+",col=color)
loc<-data.frame(loc)
colnames(loc)<-colnames
title(txt,col.main="white")
return(unlist(loc))
} 
format0<-function(x,digits=0)
{
if(digits==0) digits=1+trunc(max(log10(max(x,1))))
xx<-as.character(x)
nx<-nchar(xx)
out<-rep(paste(rep("0",digits),collapse=""),length(xx))
substring(out,digits-nx+1,digits)<-xx
return(out)
}
format0s<-function(x,left=0,digits=0,signs=c("-","",""),prefix=T)
{
signx<-signs[sign(x)+2]
x<-format00(x=x,left=left,digits=digits)
if(prefix)
return(paste(signx,x,sep=""))
else
return(paste(x,signx,sep=""))
}
format00<-function(x,left=0,digits=0)
{
sigx<-sign(x)
x<-round(abs(x),digits)
if(digits==0) return(format0(x,digits=left))
else if(digits<0) return(format0(x%/%10^(-digits),digits=left))
else return(paste(format0(x%/%1,digits=left),".",format0(round(x%%1*10^digits),digits=digits),sep=""))
}

plot.cat<-function(x,field="WmaxS",colors=c("white","black","blue","green","yellow","orange","red"),cuts=c(0,34,64,83,96,114,135,Inf),leg=c(-75,35),cnts=NULL, add.legend=F)     
{      
if(length(colors) !=length(cuts)-1) stop("Number of colors is number of cuts minus 1")
catcut<-as.numeric(cut(x[,field],cuts))
sps<-split(x[,c("lon","lat",field)],catcut)
cats<-as.numeric(names(sps))
for(i in 1:length(cats))  points2(sps[[i]],pch=".",cex=3,col=colors[cats[i]])  
cnt<-sapply(sps,nrow)
cnt0<-rep(0,length(colors))
cnt0[cats]<-cnt 
cnts<-c(cnts,list(cnt0)) 

if(add.legend)
{
names<-format(c("Tdep" ,paste("Cat",0:5," ",sep="")))
cntstxt<-lapply(cnts,function(x) paste(" ",format0(x)," ",sep=""))
txt<-paste(names," - ",do.call("paste",c(cntstxt,sep=",")),sep="")
legend(leg[1],leg[2],txt,col=colors,pch=16)
}
return(cnts)
}

points2<-function(d,...)
{
points(x=d[,1],y=d[,2],...)
}


HurRisk<-function(r=NULL,loc=NULL,iter=NULL,new=F,bbox=data.frame(lon=c(-100,-70),lat=c(25,45)),lg=landGrid, file=NULL ,field="WmaxS",velunits=c("kt","mph","k/h","m/s"), distunits=c("nm","miles","km"),umin=35,threshold=64,se=c(1851,2005),cats=c(64,83,96,114,135),rp=c(2,5,10,20,50,100,200,500,1000,10000), p=c(.025,.05,.10,.25,.5,.75,.90,.95,.975),keep.data=F)
{
  nyear<-diff(se)+1
###r, radii of interst, 
###loc, list of locations,
###Provides graphical user interface User coordinates in degrees longitude and degrees latitude with distance nautical miles and velocity in kt
###Primary purpose is to act as interface to code, output file is generated for each location and each radii, 
###Radius resolution 1 nm, lat,lon resolution .1 degrees (to match best track) unless specified by user.
###umin, minimum wind speed in kt to keep in data sets
###threshold in velunits 
###velunits one of c("kt","mph","k/h","m/s"
###runits one of c("nm","miles","km" 
###The rp.ci and rl.ci are based on the return rate e.g. 1/rp = lambda = (1+xi/sigma*(rl-mu))^(-1/xi) for return level (rl), location mu,scale sigma and shape xi.
  N<-max(length(r),nrow(loc),1)  ###loop size 
  if(N==1 && !is.null(iter))  N<-iter
  if(!is.null(loc) && !is.null(r))
  {
    ll<-nrow(loc)
    lr<-length(r)
    if( ((ll < N) && (ll!=1)) || ((lr < N) && (lr!=1))) stop("Radius and location length mismatch")
    if(lr<N) r<-rep(r,N) else
    if(ll<N) loc<-loc[rep(1,N),]
  }
        convsv<-1/c(1,0.868976242,0.539956803, 1.943844492)   ###convert kt to kt,kt to  mph, kt to k/h, kt to m/s 
        cats<-round(cats*match.unit(velunits,conv=convsv),1)

        velunits<-velunits[1]                    
        convsd<-1/c(1,0.868976242,0.539956803)   ###convert nm to nm, nm to miles, nm to km (if missing r)
        convdistin<-match.unit(distunits,conv=convsd)
        distunits<-distunits[1]

    for(i in 1:N)
    {
        if(i==1 || new) 
        {
        x11()
        plot(type="n",bbox$lon,bbox$lat,xlab="longitude",ylab="latitude")
        points2(lg,col="purple",type="l")
        map("state",add=T,resolution=0,)
        map.dev<-dev.cur()

        }
        else
        {
        if(dev.cur() != map.dev) dev.set(map.dev)

        }
 
        if(is.null(loc))
            loc1<-select.location(txt="Select Point of Interest",color="red",digits=1)
        else
        {
            loc1<-unlist(loc[i,])
            points(loc1[1],loc1[2],pch="+",col="red")
        }
        loc1xyz<-lonlattoxyz(lon=loc1[1],lat=loc1[2])
        
        ####Get and fix units    
        if(is.null(r))
        {  
           loc2<-select.location(txt="Now select a point on the search radius",color="blue",digits=NULL)
        }
        else
        {            
            loc2<-loc1+c(0,.1)
            loc3<-loc1+c(.1,0)
        }
 
        loc2xyz<-lonlattoxyz(lon=loc2[1],lat=loc2[2])
        locd1<-loc1xyz-loc2xyz
        if(is.null(r))
        {        
        radius<-round(convdistin*sqrt(sum(locd1^2)))
        loc3<-xyztolonlatheight(data=data.frame(loc1xyz+cross(loc1xyz,locd1)/sqrt(sum(loc1xyz^2))))[,c("lon","lat")]
        }
        else
        {
        radius<-r[i]
        }
        loc3xyz<-lonlattoxyz(lon=loc3[1],lat=loc3[2])
        locd1<-locd1*radius/sqrt(sum(locd1^2))/convdistin
        locd2<-loc3xyz-loc1xyz
        locd2<-locd2*radius/sqrt(sum(locd2^2))/convdistin
        w<-array(0,c(361,3))
        for(k in 1:361) w[k,]<-loc1xyz+locd1*cos(k*pi/180)+locd2*sin(k*pi/180)
        colnames(w)<-c("x","y","z")
        ww<-xyztolonlatheight(data=w)[,c("lon","lat")]
        points2(ww,col="green",type="l")
        data<-get.max(lon=loc1[1],lat=loc1[2],radius=radius,umin=umin,distunits=distunits,velunits=velunits,field=field,se=se) 
        if(i==1)  cnts<-NULL
        if(i==N)  add.legend<-T
        else add.legend<-F
        cnts<-plot.cat(data,cnts=cnts,add.legend=add.legend,leg=c(-75-2*N,33))
        if(i==N)
        title("Locations, intensities, and counts of \n tropical storm maxima within selected radii")    
        x11()   
        id<-paste("lon=",format0s(loc1[1],digits=1,signs=c("W","E","E"),prefix=F)," lat=",format0s(loc1[2],digits=1,signs=c("S","N","N"),prefix=F)," radius=",radius,distunits,sep=""," threshold=",threshold,velunits)
        data1<-exceed.pp.plot(data$y,threshold=threshold,npy=nrow(data)/nyear,addtext=id,xunit="Years",yunit=velunits)
        rp.ci<-get.conf(model=data1$model,y=cats,p=p,f="pp.rate",units=velunits,digits=2)
        rl.ci<-get.conf(model=data1$model,y=rp,p=p,f="pp.rl",units="yr",digits=1)  
        if(!keep.data) data1<-NULL                                                                                                                  
        out<-list(list(rp.ci=rp.ci,rl.ci=rl.ci,loc=loc1,radius=radius,threshold=threshold,distunits=distunits,velunits=velunits,years=se,data=data1))
        names(out)<-id
        if(i > 1)
            loctest<-c(loctest,out)
        else
            loctest<-out
    }
    if(N==1) return(loctest[[1]])
    else
return(loctest)
}

getSid<-function(lon,lat,radius,x,SidDist=NULL,se=NULL,field="WmaxS",umin=34)
{
#x is full dataframe includins Sid,
    if(is.null(SidDist))
    {
     n<-1
    }
    else
    {
     x<-x[x$Sid %in% SidDist$Sid,]
     n<-ncol(SidDist)
    }   
    xin<-inside.lonlat(x=x,lon=lon,lat=lat,yrlim=se,r=radius) & x[,field] > umin 
     if(sum(xin)==0) stop(paste("Please increase your search radius, no storms selected for",n,"iteration."))
    x<-x[xin,]
    pos<-lonlattoxyz(data=x)
    pos0<-lonlattoxyz(lon=lon,lat=lat)
    xdist<-distance.from.center(pos0=pos0,pos=pos)
    xin<- xdist <= radius
    xdist<-xdist[xin]
    if(sum(xin)==0) stop(paste("Please increase your search radius, no storms selected for",n,"iteration."))
    x<-x[xin,]
    dist<-sapply(split(xdist,x$Sid),min) #Get minimum distance
    Sid<-sapply(split(x$Sid,x$Sid),function(x) x[1]) #Get proper Sid
    if(is.null(SidDist)) 
     {
      SidDist<-data.frame(Sid=Sid,dist.1=dist)
     }
     else
     {
      SidDist<-cbind(SidDist[match(Sid,SidDist$Sid),],dist=dist)
      colnames(SidDist)[n+1]<-paste("dist",n,sep=".")
     }
 SidDist
}
    
findtracks<-function(loc=NULL,r=NULL,N=10,x=best.use,se=c(1851,2006),field="WmaxS",umin=34,bbox=data.frame(lon=c(-100,-40),lat=c(15,45)),lg=landGrid)
{
#loc is a set of locations,
#r is the search radius in nautical miles
#If loc is missing or r is missing then the user can input the data points as needed. 
#Set up polots:
        x11()
        plot(type="n",bbox$lon,bbox$lat,xlab="longitude",ylab="latitude")
        points2(lg,col="purple",type="l") ##Add best landGrid, may be removed if landGrid is not available. 
        map("state",add=T,resolution=0,)
        map(add=T) ##Get global map
        map.dev<-dev.cur()
        locations<-select.plot.points(loc=loc,r=r)
        out<- get.tracks(locations=locations,N=N,x=x,se=se,field=field,umin=umin)
        tracks<-out$tracks
        dist<-out$SidDist$Dist
        N<-out$N
        colors<-rainbow(n=N,start=.7,end=.3)
        for(i in 1:length(tracks))
        {
         xx<-tracks[[i]] 
         lines(x=xx[,"lon"],y=xx[,"lat"],col=colors[i])
        }
        
        legend(bbox$lon[2]-10,bbox$lat[2],legend=format(round(dist),0),lwd=2,col=colors, title="Distance")
        return(out)
}

select.plot.points<-function(loc,r)
{

###We plot and generate points as needed.
if(is.null(loc))
 {
 n<-as.integer(readline("How many data points do you need? "))
 if(n<=0) stop("Invalid number of data points")
 if(n>=5) warning("Many data points selected, may not return any paths")
 loc<-data.frame(lon=rep(NA,n),lat=rep(NA,n))
 getloc<-T
 } 
  else
 {
 n<-nrow(loc)
 loc<-data.frame(loc)[c("lon","lat")]
 getloc<-F
 }
 if(is.null(r))
 {
 r<-rep(NA,n)
 getr<-T
 }
 else
 {
  if(length(r)==1) r<-rep(r,n)
  else if(length(r) !=n) stop("Invalid length for radius")
  getr<-F
 }
for(i in 1:n)
{
      if(getloc)
         {
            loc1<-select.location(txt="Select Point of Interest",color="red",digits=1)
            loc[i,]<-loc1
         }
        else
        {
            loc1<-unlist(loc[i,])
            points(loc1[1],loc1[2],pch="+",col="red")
        }
        loc1xyz<-lonlattoxyz(lon=loc1[1],lat=loc1[2])
        
        ####Get     
        if(getr)
        {  
           loc2<-select.location(txt="Now select a point on maximum search radius",color="blue",digits=NULL)
        }
        else
        {            
            loc2<-loc1+c(0,.1)
            loc3<-loc1+c(.1,0)
        }
 
        loc2xyz<-lonlattoxyz(lon=loc2[1],lat=loc2[2])
        locd1<-loc1xyz-loc2xyz
        if(getr)
        {        
        radius<-round(sqrt(sum(locd1^2)))
        r[i]<-radius
        loc3<-xyztolonlatheight(data=data.frame(loc1xyz+cross(loc1xyz,locd1)/sqrt(sum(loc1xyz^2))))[,c("lon","lat")]
        }
        else
        {
        radius<-r[i]
        }
        loc3xyz<-lonlattoxyz(lon=loc3[1],lat=loc3[2])
        locd1<-locd1*radius/sqrt(sum(locd1^2))
        locd2<-loc3xyz-loc1xyz
        locd2<-locd2*radius/sqrt(sum(locd2^2))
        w<-array(0,c(361,3))
        for(k in 1:361) w[k,]<-loc1xyz+locd1*cos(k*pi/180)+locd2*sin(k*pi/180)
        colnames(w)<-c("x","y","z")
        ww<-xyztolonlatheight(data=w)[,c("lon","lat")]
        points2(ww,col="green",type="l")
        }

return(data.frame(loc,R=r))
}


get.tracks<-function(locations,N=10,x=best.use,se=c(1851,2006),field="WmaxS",umin=34)
{
#The locations is a dataframe with three columns lat, lon, R
#R is the search radius to look for tracks. It is in nautical miles.
#N is the number of tracks to find.
#x is the hurricane data file.

#Each row in the locations database is a point associated with a historical track. 
#The algorithm searches for all paths taht fall within the given radii and generates a score based
#on the square root of the sum of the euclidean distances for all storms, from the hourly interpolations.
#The output contains of a single list. The first item is a vector of distances, and the second item is a list in which each corresponding
#element is a track, for example out$dist[4] is the distance of the 4th possible track, and out$track[[4]] is the actual track from the track file
SidDist<-NULL
for(i in 1:nrow(locations))
{
SidDist<-getSid(lon=locations[i,"lon"],lat=locations[i,"lat"],radius=locations[i,"R"],SidDist=SidDist,se=se,field=field,umin=umin,x=x)
}
Sid<-SidDist$Sid
dist<-apply(SidDist[,-1,drop=F],1,function(x) sqrt(sum(x^2)))
od<-order(dist)
SidDistO<-data.frame(Sid=Sid,Dist=dist) #SidDistO, not SidDist0
SidDistO<-SidDistO[od,]
NN<-nrow(SidDistO)
if(NN>=N) SidDistO<-SidDistO[1:N,]
 else
 {
 warning(paste("We found only",NN,"tracks"))
 N<-NN
 }
 tracks<-lapply(SidDistO$Sid,function(i,xx=x) x[x$Sid==i,])
 return(list(tracks=tracks,SidDist=SidDistO,N=N,locations=locations)) 
}
